#include "terminal_user_input.h"
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

// declare an enum for the directions
typedef enum { NORTH, SOUTH, EAST, WEST } direction;

// forward declare the room pointer. The room_ptr type is a pointer to struct room_data
typedef struct room_data *room_ptr;

// a path goes from one room to another in this game
typedef struct
{
	direction 	heading;		// direction this heads
	my_string	view_text;		// text to show the user when they look at the room
	my_string	travel_text;	// text to show the user when they use this path
	room_ptr	destination;	// the room they end up in when they use this path
} path_data, *path_ptr;

typedef struct room_data
{
	my_string 	title, description;	// the title and description of the room
	path_ptr	exits;				// a pointer to dynamic memory for the exits (array)
	int 		exit_count;			// the number of exits in this room
} room_data;

typedef struct
{
	room_ptr	rooms;				// pointer to dynamic array of rooms (array)
	int 		room_count;			// the number of rooms in the array
} maze_data;

// Convert a character into a direction.
// Used when file is loaded
direction ch_to_dir(char ch)
{
	switch(ch)
	{
	  case 'n': case 'N': return NORTH;
	  case 's': case 'S': return SOUTH;
	  case 'e': case 'E': return EAST;
	  case 'w': case 'W': return WEST;
	  default: return NORTH;
	}
}

// Converts a direction into a string
// Uses my_string as you cannot return an array of characters
my_string dir_to_string(direction dir)
{
	my_string result;
	switch(dir)
	{
	  case NORTH:
	  	strncpy(result.str, "North", 6); // copy includes null
	  	break;
	  case SOUTH:
	  	strncpy(result.str, "South", 6);
	  	break;
	  case EAST:
	  	strncpy(result.str, "East", 5);
	  	break;
	  case WEST:
	  	strncpy(result.str, "West", 5);
	  	break;
	  default:
	  	strncpy(result.str, "Unknown", 8);
	}
	return result;
}

// Adds a path from one room to another.
// - from_room = 	pointer to the room it heads from
// - to_room = 		pointer to the room it heads to
// - head = 		direction of path
// - view_text = 	text to show user when they view the path
// - travel_text = 	text to show user when they use this path
void add_path(room_ptr from_room, room_ptr to_room, direction head, my_string view_text, my_string travel_text)
{
	int last_idx;
	path_ptr new_exits;

	// increase number of exits
	from_room->exit_count++;

	// TODO: 2 -> need to call realloc here!
	//
	// -- change the allocation of from_room->exits
	// -- new size needs to store 'from_room->exit_count' number of path_data values
	new_exits = NULL; // need to use realloc to allocate space!

	if (new_exits == NULL)
	{
		printf("Error allocating space in add_path -> either out of memory or realloc call not added.");
		exit(EXIT_FAILURE);
	}

	from_room->exits = new_exits;

	last_idx = from_room->exit_count - 1;
	from_room->exits[last_idx].destination 	= to_room;
	from_room->exits[last_idx].heading 		= head;
	from_room->exits[last_idx].view_text 	= view_text;
	from_room->exits[last_idx].travel_text 	= travel_text;
}

//
// Loads the maze from file.
//
void load_maze(const char *filename, maze_data *maze)
{
	FILE *input;
	char dir;
	int room_count, exit_count, from_room_idx, to_room_idx;
	int room_data_size;
	int i;
	my_string view_text, travel_text;
	direction head;

	input = fopen(filename, "r");
	if ( input == NULL )
	{
		printf("Unable to open maze file.");
		exit(EXIT_FAILURE);
	}

	// Read the number of rooms
	fscanf(input, "%d ", &room_count);
	maze->room_count = room_count;

	// Allocate space for rooms in memory
	room_data_size = sizeof(room_data) * room_count;
	maze->rooms = (room_ptr) malloc(room_data_size);
	// Clear the space so it is all full of 0s
	memset(maze->rooms, 0, room_data_size);

	// Loop for each room and read it from the file
	for (i = 0; i < room_count; i++)
	{
		// printf("Loading room %d\n", i);
		fscanf(input, "%255[^\n] ", maze->rooms[i].title.str);
		fscanf(input, "%255[^\n] ", maze->rooms[i].description.str);
		maze->rooms[i].exit_count = 0;
		maze->rooms[i].exits = NULL;
	}

	fscanf(input, "%d ", &exit_count);
	// printf("%d\n", exit_count);

	for (i = 0; i < exit_count; i++)
	{
		fscanf(input, "%d %d %c ", &from_room_idx, &to_room_idx, &dir);
		fscanf(input, "%255[^\n] ", view_text.str);
		fscanf(input, "%255[^\n] ", travel_text.str);

		printf("Adding a door from %d (index %d) to %d (index %d) -> %c\n",
			from_room_idx, 		// values in the file are 1 based...
			from_room_idx - 1, 	// index are 0 based so -1 to get array index
			to_room_idx,
			to_room_idx - 1, 	// same for the to_room_idx
			dir);
		printf(" from room is -> %s\n", maze->rooms[from_room_idx - 1].title.str);

		// TODO: 1 -> need to call add_path here!
		//
		// add_path(??,	// pointer to from room (use from_room_idx - 1)
		// 			??, // pointer to to room
		// 			ch_to_dir(dir), // head direction
		// 			view_text, 		// view text
		// 			travel_text );	// travel text
	}

	fclose(input);
}

//
// shows the room to the user
//
void show_room(room_ptr room)
{
	my_string heading;
	int i;

	printf("----- %s -----\n", room->title.str);
	printf("%s\n\n", room->description.str);

	printf("There are exits heading...\n");

	for (i = 0; i < room->exit_count; i++)
	{
		heading = dir_to_string(room->exits[i].heading);
		printf("\t%d: [%s] %s\n",
				i + 1,
				heading.str,
				room->exits[i].view_text.str);
	}
	printf("\n");
}

//
// entry point - loops the game until user enter 0
//
int main()
{
	maze_data my_maze;
	room_ptr player_room;
	int option;

	load_maze("maze.txt", &my_maze);

	player_room = &my_maze.rooms[0];

	do
	{
		// show the player the room
		show_room(player_room);

		// get their exit choice
		option = read_integer_range("Take exit (0 quit): ", 0, player_room->exit_count) - 1;

		if ( option >= 0 )
		{
			// print the travel text
			printf("\n%s\n\n", player_room->exits[option].travel_text.str);

			// TODO: 3 - move the player
			//
			// need to change the player_room...
			//   get the selected (using option) exit from the player_room (the path they chose)
			//   get the destination from this path
			//   assign this to the player_room
			// 	 (so player is now in the destination room of the selected path from the current room)
			//
			// player_room = ??;
		}
	} while ( option != -1 );
}
